/*
 * The contents of this file are subject to the terms
 * of the Common Development and Distribution License
 * (the License).  You may not use this file except in
 * compliance with the License.
 *
 * You can obtain a copy of the license at
 * https://glassfish.dev.java.net/public/CDDLv1.0.html or
 * glassfish/bootstrap/legal/CDDLv1.0.txt.
 * See the License for the specific language governing
 * permissions and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL
 * Header Notice in each file and include the License file
 * at glassfish/bootstrap/legal/CDDLv1.0.txt.
 * If applicable, add the following below the CDDL Header,
 * with the fields enclosed by brackets [] replaced by
 * you own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 */

package com.sun.grizzly;

import com.sun.grizzly.util.Copyable;
import java.io.IOException;
import java.net.BindException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.net.SocketException;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.Iterator;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * A SelectorHandler handles all java.nio.channels.Selector operations.
 * One or more instance of a Selector are handled by SelectorHandler.
 * The logic for processing of SelectionKey interest (OP_ACCEPT,OP_READ, etc.)
 * is usually defined using an instance of SelectorHandler.
 *
 * This class represents a TCP implementation of a SelectorHandler.
 * This class first bind a ServerSocketChannel to a TCP port and then start
 * waiting for NIO events.
 *
 * @author Jeanfrancois Arcand
 */
public class TCPSelectorHandler implements SelectorHandler {
    
    
    /**
     * The ConnectorInstanceHandler used to return a new or pooled
     * ConnectorHandler
     */
    protected ConnectorInstanceHandler connectorInstanceHandler;
    
    
    /**
     * The list of SelectionKey to register next time the Selector.select is
     * invoked.
     */
    protected ConcurrentLinkedQueue<SelectionKey> opReadToRegister;
    
    
    /**
     * The list of SelectionKey to register next time the Selector.select is
     * invoked.
     */
    protected ConcurrentHashMap<SocketAddress[],CallbackHandler>
            opConnectToRegister;
    
    
    /**
     * The list of SelectionKey to register next time the Selector.select is
     * invoked.
     */
    protected ConcurrentLinkedQueue<SelectionKey> opWriteToRegister;
    
    
    /**
     * The socket tcpDelay.
     */
    protected boolean tcpNoDelay = false;
    
    
    /**
     * The socket reuseAddress
     */
    protected boolean reuseAddress = true;
    
    
    /**
     * The socket linger.
     */
    protected int linger = 100;
    
    /**
     * The socket time out
     */
    protected int socketTimeout = -1;
    
    
    protected Logger logger;
    
    
    /**
     * The server socket time out
     */
    protected int serverTimeout = 0;
    
    
    /**
     * The inet address to use when binding.
     */
    protected InetAddress inet;
    
    
    /**
     * The default TCP port.
     */
    protected int port = 18888;
    
    
    /**
     * The ServerSocket instance.
     */
    protected ServerSocket serverSocket;
    
    
    /**
     * The ServerSocketChannel.
     */
    protected ServerSocketChannel serverSocketChannel;
    
    
    /**
     * The single Selector.
     */
    protected Selector selector;
    
    
    /**
     * The Selector time out.
     */
    protected long selectTimeout = 1000L;
    
    
    /**
     * Server socket backlog.
     */
    protected int ssBackLog = 4096;
    
    
    /**
     * Is this used for client only or client/server operation.
     */
    protected boolean isClient = false;
    
    
    public TCPSelectorHandler(){
    }
    
    
    public TCPSelectorHandler(boolean isClient) {
        this.isClient = isClient;
    }
    
    
    public void copyTo(Copyable copy) {
        TCPSelectorHandler copyHandler = (TCPSelectorHandler) copy;
        copyHandler.selector = selector;
        copyHandler.selectTimeout = selectTimeout;
        copyHandler.serverTimeout = serverTimeout;
        copyHandler.inet = inet;
        copyHandler.port = port;
        copyHandler.ssBackLog = ssBackLog;
        copyHandler.tcpNoDelay = tcpNoDelay;
        copyHandler.linger = linger;
        copyHandler.socketTimeout = socketTimeout;
        copyHandler.logger = logger;
        copyHandler.reuseAddress = reuseAddress;
        copyHandler.connectorInstanceHandler = connectorInstanceHandler;
    }
    
    
    /**
     * Return the set of SelectionKey registered on this Selector.
     */
    public Set<SelectionKey> keys(){
        if (selector != null){
            return selector.keys();
        } else {
            throw new IllegalStateException("Selector is not created!");
        }
    }
    
    
    /**
     * Is the Selector open.
     */
    public boolean isOpen(){
        if (selector != null){
            return selector.isOpen();
        } else {
            return false;
        }
    }
    
    
    /**
     * Before invoking Selector.select(), make sure the ServerScoketChannel
     * has been created. If true, then register all SelectionKey to the Selector.
     * @param ctx <code>Context</code>
     */
    public void preSelect(Context ctx) throws IOException {
        if (selector == null){
            try{
                connectorInstanceHandler = new TCPConnectorInstanceHandler();
                
                // Create the socket listener
                selector = Selector.open();
                
                if (!isClient){
                    serverSocketChannel = ServerSocketChannel.open();
                    serverSocket = serverSocketChannel.socket();
                    serverSocket.setReuseAddress(reuseAddress);
                    if ( inet == null)
                        serverSocket.bind(new InetSocketAddress(port),ssBackLog);
                    else
                        serverSocket.bind(new InetSocketAddress(inet,port),ssBackLog);
                    
                    serverSocketChannel.configureBlocking(false);
                    serverSocketChannel.register(selector, SelectionKey.OP_ACCEPT);
                }
            } catch (SocketException ex){
                throw new BindException(ex.getMessage() + ": " + port + "=" + this);
            }
            
            if (!isClient){
                serverSocket.setSoTimeout(serverTimeout);
            }
        } else {
            if (opReadToRegister == null){
                opReadToRegister = new ConcurrentLinkedQueue<SelectionKey>();
            }
            
            SelectionKey key;
            int size = opReadToRegister.size();
            long currentTime = 0L;
            if (size > 0){
                currentTime = System.currentTimeMillis();
            }
            
            for (int i=0; i < size; i++) {
                key = opReadToRegister.poll();
                
                if (!key.isValid()){
                    continue;
                }
                
                key.interestOps(key.interestOps() | SelectionKey.OP_READ);
                
                if (key.attachment() == null)
                    key.attach(currentTime);
            }
            
            if (opWriteToRegister== null){
                opWriteToRegister = new ConcurrentLinkedQueue<SelectionKey>();
            }
            
            size = opWriteToRegister.size();
            currentTime = 0L;
            if (size > 0){
                currentTime = System.currentTimeMillis();
            }
            
            for (int i=0; i < size; i++) {
                key = opWriteToRegister.poll();
                
                if (!key.isValid()){
                    continue;
                }
                
                key.interestOps(key.interestOps() | SelectionKey.OP_WRITE);
                
                if (key.attachment() == null)
                    key.attach(currentTime);
            }
            
            if (opConnectToRegister== null){
                opConnectToRegister = new ConcurrentHashMap
                        <SocketAddress[],CallbackHandler>();
            }
            
            size = opConnectToRegister.size();
            currentTime = 0L;
            if (size > 0){
                currentTime = System.currentTimeMillis();
            } else {
                return;
            }
            
            CallbackHandler<SocketChannel> callbackHandler;
            SocketChannel socketChannel;
            Iterator<SocketAddress[]> iterator =
                    opConnectToRegister.keySet().iterator();
            SocketAddress[] remoteLocal;
            while(iterator.hasNext()){
                remoteLocal = iterator.next();
                socketChannel= SocketChannel.open();
                socketChannel.socket().setReuseAddress(true);
                if (remoteLocal[1] != null){
                    socketChannel.socket().bind(remoteLocal[1]);
                }
                socketChannel.configureBlocking(false);
                boolean isConnected = socketChannel.connect(remoteLocal[0]);
                key = socketChannel.register(selector,
                        SelectionKey.OP_CONNECT,opConnectToRegister.remove(remoteLocal));
                
                // if channel was connected immediately
                if (isConnected) {
                    onConnectInterest(key, ctx);
                }
            }
        }
    }
    
    
    /**
     * Execute the Selector.select(...) operations.
     * @param ctx <code>Context</code>
     * @return <code>Set</code> of <code>Context</code>
     */
    public Set<SelectionKey> select(Context ctx)
            throws IOException{
        
        selector.select(selectTimeout);
        return selector.selectedKeys();
    }
    
    
    /**
     * Invoked after Selector.select().
     * NOT USED.
     * @param ctx <code>Context</code>
     */
    public void postSelect(Context ctx) {
    }
    
    
    /**
     * Register a SelectionKey to this Selector.
     */
    public void register(SelectionKey key, int ops) {
        if (ops == SelectionKey.OP_READ){
            opReadToRegister.offer(key);
        } else if (ops == SelectionKey.OP_WRITE){
            opWriteToRegister.offer(key);
        } else if (ops == (SelectionKey.OP_WRITE|SelectionKey.OP_READ)){
            opReadToRegister.offer(key);
            opWriteToRegister.offer(key);
        }
        selector.wakeup();
    }
    
    
    /**
     * Register a CallBackHandler to this Selector.
     *
     * @param remoteAddress remote address to connect
     * @param localAddress local address to bin
     * @param callBackHandler <code>CallbackHandler</code>
     * @throws java.io.IOException
     */
    protected void connect(SocketAddress remoteAddress, SocketAddress localAddress,
            CallbackHandler callBackHandler) throws IOException{
        
        if (opConnectToRegister== null){
            opConnectToRegister = new ConcurrentHashMap
                    <SocketAddress[],CallbackHandler>();
        }
        
        opConnectToRegister.put(new SocketAddress[]{remoteAddress,localAddress},
                callBackHandler);
        selector.wakeup();
    }
    
    
    /**
     * Shuntdown this instance by closing its Selector and associated channels.
     */
    public void shutdown(){
        try{
            if (serverSocket != null)
                serverSocket.close();
        } catch (Throwable ex){
            Controller.logger().log(Level.SEVERE,
                    "serverSocket.close",ex);
        }
        
        try{
            if (serverSocketChannel != null)
                serverSocketChannel.close();
        } catch (Throwable ex){
            Controller.logger().log(Level.SEVERE,
                    "serverSocketChannel.close",ex);
        }
        
        try{
            if (selector != null)
                selector.close();
        } catch (Throwable ex){
            Controller.logger().log(Level.SEVERE,
                    "selector.close",ex);
        }
        
    }
    
    /**
     * {@inheritDoc}
     */
    public SelectableChannel acceptWithoutRegistration(SelectionKey key)
            throws IOException {
        ServerSocketChannel server = (ServerSocketChannel) key.channel();
        SocketChannel channel = server.accept();
        return channel;
    }
    
    /**
     * Handle OP_ACCEPT.
     * @param ctx <code>Context</code>
     * @return always returns false
     */
    public boolean onAcceptInterest(SelectionKey key,
            Context ctx) throws IOException{
        SelectableChannel channel = acceptWithoutRegistration(key);
        
        if (channel != null) {
            configureChannel(channel);
            SelectionKey readKey =
                    channel.register(selector, SelectionKey.OP_READ);
            readKey.attach(System.currentTimeMillis());
        }
        return false;
    }
    
    /**
     * Handle OP_READ.
     * @param ctx <code>Context</code>
     * @param key <code>SelectionKey</code>
     * @return false if handled by a <code>CallbackHandler</code>, otherwise true
     */
    public boolean onReadInterest(final SelectionKey key,final Context ctx)
            throws IOException{
        // disable OP_READ on key before doing anything else
        key.interestOps(key.interestOps() & (~SelectionKey.OP_READ));
        if (key.attachment() instanceof CallbackHandler){
            final Context context = ctx.getController().pollContext(key);
            context.setSelectionKey(key);
            context.setCurrentOpType(Context.OpType.OP_READ);
            invokeCallbackHandler(context);
            return false;
        } else {
            return true;
        }
    }
    
    
    /**
     * Handle OP_WRITE.
     *
     * @param key <code>SelectionKey</code>
     * @param ctx <code>Context</code>
     */
    public boolean onWriteInterest(final SelectionKey key,final Context ctx)
            throws IOException{
        // disable OP_WRITE on key before doing anything else
        key.interestOps(key.interestOps() & (~SelectionKey.OP_WRITE));
        if (key.attachment() instanceof CallbackHandler){
            final Context context = ctx.getController().pollContext(key);
            context.setSelectionKey(key);
            context.setCurrentOpType(Context.OpType.OP_WRITE);
            invokeCallbackHandler(context);
            return false;
        } else {
            return true;
        }
    }
    
    
    /**
     * Handle OP_CONNECT.
     * @param key <code>SelectionKey</code>
     * @param ctx <code>Context</code>
     */
    public boolean onConnectInterest(final SelectionKey key, Context ctx)
            throws IOException{
        // disable OP_CONNECT on key before doing anything else
        key.interestOps(key.interestOps() & (~SelectionKey.OP_CONNECT));
        
        // No OP_READ nor OP_WRITE allowed yet.
        key.interestOps(key.interestOps() & (~SelectionKey.OP_WRITE));
        key.interestOps(key.interestOps() & (~SelectionKey.OP_READ));
        
        if (key.attachment() instanceof CallbackHandler){
            Context context = ctx.getController().pollContext(key);
            context.setSelectionKey(key);
            context.setCurrentOpType(Context.OpType.OP_CONNECT);
            invokeCallbackHandler(context);
        }
        return false;
    }
    
    
    /**
     * Invoke a CallbackHandler via a Context instance.
     * @param context <code>Context</code>
     * @throws java.io.IOException
     */
    protected void invokeCallbackHandler(Context context) throws IOException{
        context.setProtocol(protocol());
        
        IOEvent<Context>ioEvent = createIOEvent();
        ioEvent.attach(context);
        context.setIOEvent(ioEvent);
        try {
            context.execute();
        } catch (PipelineFullException ex){
            throw new IOException(ex.getMessage());
        }
    }
    
    
    /**
     * Return an instance of the default <code>ConnectorHandler</code>,
     * which is the <code>TCPConnectorHandler</code>
     * @return <code>ConnectorHandler</code>
     */
    public ConnectorHandler acquireConnectorHandler(){
        if (selector == null && !selector.isOpen()){
            throw new IllegalStateException("SelectorHandler not yet started");
        }
        
        ConnectorHandler connectorHandler = connectorInstanceHandler.acquire();
        return connectorHandler;
    }
    
    
    /**
     * Release a ConnectorHandler.
     */
    public void releaseConnectorHandler(ConnectorHandler connectorHandler){
        connectorInstanceHandler.release(connectorHandler);
    }
    
    
    /**
     * A token decribing the protocol supported by an implementation of this
     * interface
     */
    public Controller.Protocol protocol(){
        return Controller.Protocol.TCP;
    }
    // ------------------------------------------------------ Utils ----------//
    
    
    /**
     * {@inheritDoc}
     */
    public void configureChannel(SelectableChannel channel) throws IOException{
        Socket socket = ((SocketChannel) channel).socket();
        
        channel.configureBlocking(false);
        
        try{
            if(linger >= 0 ) {
                socket.setSoLinger( true, linger);
            }
        } catch (SocketException ex){
            logger.log(Level.WARNING,
                    "setSoLinger exception ",ex);
        }
        
        try{
            if(tcpNoDelay)
                socket.setTcpNoDelay(tcpNoDelay);
        } catch (SocketException ex){
            logger.log(Level.WARNING,
                    "setTcpNoDelay exception ",ex);
        }
        
        try{
            socket.setReuseAddress(reuseAddress);
        } catch (SocketException ex){
            logger.log(Level.WARNING,
                    "setReuseAddress exception ",ex);
        }
    }
    
    
    // ------------------------------------------------------ Properties -----//
    
    public final Selector getSelector() {
        return selector;
    }
    
    public final void setSelector(Selector selector) {
        this.selector = selector;
    }
    
    public long getSelectTimeout() {
        return selectTimeout;
    }
    
    public void setSelectTimeout(long selectTimeout) {
        this.selectTimeout = selectTimeout;
    }
    
    public int getServerTimeout() {
        return serverTimeout;
    }
    
    public void setServerTimeout(int serverTimeout) {
        this.serverTimeout = serverTimeout;
    }
    
    public InetAddress getInet() {
        return inet;
    }
    
    public void setInet(InetAddress inet) {
        this.inet = inet;
    }
    
    public int getPort() {
        return port;
    }
    
    public void setPort(int port) {
        this.port = port;
    }
    
    public int getSsBackLog() {
        return ssBackLog;
    }
    
    public void setSsBackLog(int ssBackLog) {
        this.ssBackLog = ssBackLog;
    }
    
    
    public boolean isTcpNoDelay() {
        return tcpNoDelay;
    }
    
    public void setTcpNoDelay(boolean tcpNoDelay) {
        this.tcpNoDelay = tcpNoDelay;
    }
    
    public int getLinger() {
        return linger;
    }
    
    public void setLinger(int linger) {
        this.linger = linger;
    }
    
    public int getSocketTimeout() {
        return socketTimeout;
    }
    
    public void setSocketTimeout(int socketTimeout) {
        this.socketTimeout = socketTimeout;
    }
    
    public Logger getLogger() {
        return logger;
    }
    
    public void setLogger(Logger logger) {
        this.logger = logger;
    }
    
    public boolean isReuseAddress() {
        return reuseAddress;
    }
    
    public void setReuseAddress(boolean reuseAddress) {
        this.reuseAddress = reuseAddress;
    }
    
    /**
     * Create IOEvent instance
     * @return IOEvent
     */
    protected static <E> IOEvent<E> createIOEvent() {
        return new IOEvent<E>() {
            private E context;
            public E attach(E context){
                this.context = context;
                return context;
            }
            
            public E attachment(){
                return context;
            }
        };
        
    }
}
